jp_DungeonGenerator
	var
		const
			FALSE = 0
			TRUE = 1

		turf/corner1 //One corner of the rectangle the algorithm is allowed to modify
		turf/corner2 //The other corner of the rectangle the algorithm is allowed to modify

		list/allowedRooms //The list of rooms the algorithm may place

		doAccurateRoomPlacementCheck = FALSE //Whether the algorithm should just use AABB collision detection between rooms, or use the slower version with no false positives
		usePrexistingRegions = FALSE //Whether the algorithm should find any already extant open regions in the area it is working on, and incorporate them into the dungeon being generated

		floortype //The type used for open floors placed in corridors
		list/walltype //Either a single type, or a list of types that are considered 'walls' for the purpose of this algorithm

		numRooms = 20
		numExtraPaths = 20
		maximumIterations = 20
		roomMinSize = 4
		roomMaxSize = 4
		maxPathLength = 50
		minPathLength = 2
		minLongPathLength = 15
		pathEndChance = 10
		longPathChance = 50

		list
			border_turfs

		long_counter = 0

	proc
		getAdjacent(turf/t)
			return list(get_step(t, NORTH), get_step(t, SOUTH), get_step(t, EAST), get_step(t, WEST))

		isWall(turf/t)
			return t.type in walltype

		islist(l)
			return istype(l, /list)

		setArea(turf/c1, turf/c2)
			corner1 = c1
			corner2 = c2

		getArea()
			return list(corner1, corner2)

		getMinX()
			return min(corner1.x, corner2.x)

		getMaxX()
			return max(corner1.x, corner2.x)

		getMinY()
			return min(corner1.y, corner2.y)

		getMaxY()
			return max(corner1.y, corner2.y)


		setAllowedRooms(list/l)
			allowedRooms = l

		addAllowedRoom(jp_DungeonRoom/r)
			if(!allowedRooms) allowedRooms = list()
			allowedRooms+=r

		removeAllowedRoom(jp_DungeonRoom/r)
			allowedRooms-=r
			if(!allowedRooms || !allowedRooms.len) allowedRooms = null

		getAllowedRooms()
			return allowedRooms


		setDoAccurateRoomPlacementCheck(b)
			doAccurateRoomPlacementCheck = b

		getDoAccurateRoomPlacementCheck()
			return doAccurateRoomPlacementCheck


		setUsePrexistingRegions(b)
			usePrexistingRegions = b

		getUsePrexistingRegions()
			return usePrexistingRegions


		setFloorType(f)
			floortype = f

		getFloorType()
			return floortype

		setWallType(w)
			walltype = w

		addWallType(w)
			if(!walltype) walltype = list()
			if(!islist(walltype)) walltype = list(walltype)
			walltype+=w

		removeWallType(w)
			if(!islist(walltype)) return
			walltype-=w

		getWallType()
			return walltype

		generate()
			var
				numits
				paths
				jp_DungeonRoom/nextroom
				list/jp_DungeonRoom/rooms = list()
				list/jp_DungeonRegion/regions = list()
				turf/nextloc

				minx
				maxx
				miny
				maxy
				z

				jp_DungeonRegion/region1
				jp_DungeonRegion/region2
				timer = world.timeofday

			z = corner1.z
			minx = min(corner1.x, corner2.x) + roomMaxSize
			maxx = max(corner1.x, corner2.x) - roomMaxSize
			miny = min(corner1.y, corner2.y) + roomMaxSize
			maxy = max(corner1.y, corner2.y) - roomMinSize

			while(rooms.len<numRooms)
				if(numits>maximumIterations)
					break

				nextloc = locate(rand(minx, maxx), rand(miny, maxy), z)

				nextroom = pick(allowedRooms)
				nextroom = new nextroom(rand(roomMinSize, roomMaxSize), nextloc, src)
				numits++
				if(!rooms || !intersects(nextroom, rooms))
					nextroom.place()
					numits=0
					rooms+=nextroom

			border_turfs = list()

			for(var/jp_DungeonRoom/r in rooms)
				var/jp_DungeonRegion/reg = new /jp_DungeonRegion(src)
				reg.addTurfs(r.getTurfs())
				regions+=reg
				border_turfs+=reg.getBorder()

			numits = 0
			paths = numExtraPaths
			var/temp = 0

			while(regions.len>1 || paths>0)
				if(!temp && regions.len<=1)
					temp=1
					long_counter=0

				if(numits>maximumIterations && !(regions.len>1)) break
				if(regions.len<=1) numits++
				region1 = pick(regions)
				region2 = pick(regions)

				if(region1==region2) if(regions.len>1) continue

				var/list/turf/path = getPath(region1, region2, regions)

				if(!path || !path.len) continue

				numits = 0

				if(region1==region2) if(regions.len<=1) paths--

				for(var/turf/t in path)
					path-=t
					path+=new floortype(t)

				region1.addTurfs(path)

				if(region1!=region2)
					region1.addTurfs(region2.getTurfs())
					regions-=region2

				for(var/turf/t in region1.getBorder()) if(!(t in border_turfs)) border_turfs+=t

			world << "Took [(world.timeofday - timer)/10] seconds"
			world << "Placed [numExtraPaths-paths] paths, of which [long_counter] were long"
			world << "Numits: [numits]"

		intersects(jp_DungeonRoom/newroom, list/jp_DungeonRoom/rooms)
			for(var/jp_DungeonRoom/r in rooms)
				. = newroom.getSize() + r.getSize()+1
				if((. > abs(newroom.getX() - r.getX())) && (. > abs(newroom.getY() - r.getY()))) return TRUE
			return FALSE

		getPath(jp_DungeonRegion/region1, jp_DungeonRegion/region2, list/jp_DungeonRegion/regions)
			var/turf/start = pick(region1.getBorder())
			var/turf/end
			var/minlength = minPathLength
			if(prob(longPathChance)) minlength=minLongPathLength

			var/list/borders=list()
			borders.Add(border_turfs)
			borders.Remove(region2.getBorder())

			borders-=start

			var/list/turf/previous = list()
			var/list/turf/done = list(start)
			var/list/turf/next = getAdjacent(start)
			var/list/turf/cost = list("\ref[start]"=0)

			next-=borders
			for(var/turf/t in next)
				if(!isWall(t)) next-=t
				previous["\ref[t]"] = start
				cost["\ref[t]"]=1
			if(!next.len) return list()

			while(1)
				var/turf/min
				var/mincost = maxPathLength

				for(var/turf/t in next)
					if((cost["\ref[t]"]<mincost) || (prob(50) && cost["\ref[t]"]==mincost))
						min = t
						mincost=cost["\ref[t]"]

				if(!min) return list()

				done += min
				next -= min

				if(min in region2.getBorder())
					if(mincost>minlength && prob(pathEndChance))
						if(minlength == minLongPathLength) long_counter++
						end = min
						break
					else
						continue

				for(var/turf/t in getAdjacent(min))
					if(isWall(t) && !(t in borders))
						if(!(t in done) && !(t in next))
							next+=t
							previous["\ref[t]"] = min
							cost["\ref[t]"] = mincost+1

			var/list/ret = list(end)
			var/turf/last = end
			while(1)
				if(last==start) break
				ret+=previous["\ref[last]"]
				last=previous["\ref[last]"]

			return ret

jp_DungeonRegion
	var
		jp_DungeonGenerator/gen
		list/turf/contained = list()
		list/turf/border = list()

	New(jp_DungeonGenerator/g)
		gen = g

	proc
		addTurfs(list/turf/l)
			for(var/turf/t in l)
				if(t in border) border-=t
				if(!(t in contained))
					contained+=t
					for(var/turf/t2 in gen.getAdjacent(t))
						if(gen.isWall(t2) && !(t2 in border)) border+=t2

		getTurfs()
			return contained

		getBorder()
			return border

turf
	var/temp_state

jp_DungeonRoom
	var
		turf/centre
		size
		jp_DungeonGenerator/gen

	New(s, turf/c, jp_DungeonGenerator/g)
		size = s
		centre = c
		gen = g

	proc
		getX()
			return centre.x

		getY()
			return centre.y

		getZ()
			return centre.z

		getSize()
			return size

		place()
			return

		getTurfs()
			return list()

	square
		var
			list
				turfs = list()

		getTurfs()
			return turfs

		place()
			var/upperx = round(getX() + size/2)
			var/uppery = round(getY() + size/2)

			for(var/i = upperx - size, i<upperx, i++)
				if(i<gen.getMinX() || i>gen.getMaxX()) continue
				for(var/j = uppery - size, j<uppery, j++)
					if(j<gen.getMinY() || j>gen.getMaxY()) continue
					turfs += new gen.floortype(locate(i, j, getZ()))